/* jshint bitwise:false */

/**
 * Lazy is used to do lazy loading of content.
 * It retrieve data (JSON by default) from server and injects it through a template.
 *
 * How to use it:
 * <div class="lazy" data-src="{URL to retrieve JSON}" data-tmpl="{template ID}"></div>
 *
 * Examples:
 * <div
 *  class="lazy"
 *  data-src="http://www.tomsguide.fr/community/followed/id-123456"
 *  data-tmpl="myTemplate"></div>
 * <script type="text/html" id="myTemplate"><%=foo%></script>
 *
 * If URL returns following JSON : { "foo": "bar" }, the results will be:
 * <div
 *  class="lazy"
 *  data-src="http://www.tomsguide.fr/community/followed/id-123456"
 *  data-tmpl="myTemplate">bar</div>
 * <script type="text/html" id="myTemplate"><%=foo%></script>
 *
 * For HTML injection:
 * <div
 *  class="lazy"
 *  data-src="http://www.tomsguide.fr/community/followed/id-123456"></div>
 *
 * Add your custom callback calls on your placeholder like this:
 *
 * <script type="text/javascript">
 * new Lazy({
 *    onComplete: function() { },
 *    onError: function() { }
 *    elementsAccessor: '.myPlaceholder',
 *    offsetY: 50  // (optional) vertical offset to apply to elements appearance
 *    offsetX: 0   // (optional) horizontal offset to apply to elements appearance
 * );
 * </script>
 *
 * For Javascript Lazy load :
 * <div class="lazy" data-type="script" src="http://url.to.a.script.js"></div>
 *
 * Documentation of original BOM script: http://twiki.bestofmedia.com/bin/view/Product/FELazy
 * Jasmine tests: Tests/js/modules/Purch/Utils/Lazy.spec.js
 */


define('Purch/Utils/Lazy', ['jquery', 'Purch/Utils/Templating'], function (jQuery, Templating) {

    "use strict";

    /**
     * @constructor
     * @param {Object} [oConf] configuration object
     * @param {Boolean} [oConf.init] initialization after instanciation
     * @param {String} [oConf.elementsAccessor] accessor for containers
     * @returns {Lazy} a new instance of Lazy object
     */
    var Lazy = function (oConf) {
        oConf = jQuery.extend({
            init: true,
            elementsAccessor: '.lazy',
            timer : 0,
            offsetY: 0,
            offsetX: 0,
            refreshLoaded: false,
            canReloadData: true,
            delegateOn : false,
            needHorizontal: false,
            listener: "scroll",
            onAppear: function(elt){ return true;},
            lazyEventName: 'Lazy.injectdata',
            appearEventName: 'appear',
            callbackType: "callback",
            timeout: 0
        }, oConf || {});

        /**
         * Self object
         *
         * @type Lazy
         */
        var self = this;

        /**
         * Configuration object
         *
         * @name Lazy#oConf
         * @type Object
         */
        self.oConf = oConf;

        /**
         * jQuery collection of container elements
         *
         * @name Lazy#elements
         * @type jQuery[]
         */
        self.elements = jQuery(oConf.elementsAccessor);

        /**
         * Init function to bind scroll event on window and appear event on elements
         * with the specified accessor (.lazy by default).
         *
         * The init function lauches the .scroll() function to display blocks in the
         * window on page load
         *
         * @name Lazy#init
         * @function
         */
        self.init = function () {
            if (oConf.listener === "scroll"){
                jQuery(window).on(oConf.listener,
                    function() {
                        clearTimeout(self.oConf.timer);
                        self.oConf.timer = setTimeout(function() {
                            // do something
                            self.scroll();
                        }, 10);
                    });
            }else{
                jQuery(window).on(oConf.listener, self.scroll());
            }

            if(oConf.delegateOn !== false && (typeof jQuery.fn.on === 'function'))
            {
                jQuery(oConf.delegateOn).on(oConf.appearEventName, oConf.elementsAccessor, self.retrieveData);
            }
            else
            {
                jQuery(oConf.elementsAccessor).on(oConf.appearEventName, self.retrieveData);
            }

            self.scroll();
        };

        /**
         * Refreshes collection of container elements
         * @name Lazy#refresh
         * @type jQuery[]
         */
        self.refresh = function () {
            self.elements = jQuery(oConf.elementsAccessor);
        };

        /**
         * Refreshes collection of unloaded container elements
         * @name Lazy#refresh
         * @type jQuery[]
         */
        self.refreshUnloaded = function () {
            self.elements = jQuery(oConf.elementsAccessor).not('.loaded');
        };

        /**
         * Scroll handler.
         * It loops on the elements to check if they have to appear.
         * If an element must appear, an event 'appear' is triggered.
         *
         * @name Lazy#scroll
         * @function
         */
        self.scroll = function () {
            if(self.oConf.refreshLoaded){
                self.refresh();
            }else{
                self.refreshUnloaded();
            }
            self.elements.each(function (i, elem) {
                if (self.isInWindow(elem)) {
                    jQuery(elem).trigger(oConf.appearEventName);
                }
            });
        };

        /**
         * Function to check if an element must appear because it is in the window.
         *
         * @name Lazy#isInWindow
         * @function
         * @param {jQuery} elem a jQuery DOM element
         * @returns {Boolean}
         */
        self.isInWindow = function (elem) {

            return (self.topOfElementIsInWindow(elem) || self.bottomOfElementIsInWindow(elem) || self.elementJutsOutOnTopAndBottom(elem)) &&
            (!self.oConf.needHorizontal || self.leftOfElementIsInWindow(elem) || self.rightOfElementIsInWindow(elem) || self.elementJutsOutOnLeftAndRight(elem));
        };

        self.isInWindowMiddleTop = function (elem) {
            return self.isInWindow(elem) && ( self.getElementY(elem) < (self.getWindowY() + self.getWindowHeight()/2) );
        };

        /**
         * Function to check if top border of element is in the window
         *
         * @name Lazy#topOfElementIsInWindow
         * @function
         * @param {jQuery} elem a jQuery DOM element
         * @returns {Boolean}
         */
        self.topOfElementIsInWindow = function (elem) {
            var elementY = self.getElementY(elem),
                offsetY  = self.oConf.offsetY,
                windowY = self.getWindowY(),
                windowHeight = self.getWindowHeight();

            return elementY >= windowY &&
            elementY - offsetY < windowY + windowHeight;
        };
        /**
         * Function to check if top border of element is in the window middle top
         *
         * @name Lazy#topOfElementIsInWindowMiddleTop
         * @function
         * @param {jQuery} elem a jQuery DOM element
         * @returns {Boolean}
         */
        self.topOfElementIsInWindowMiddleTop = function (elem) {
            var elementY = self.getElementY(elem),
                offsetY  = self.oConf.offsetY,
                windowY = self.getWindowY(),
                windowHeight = self.getWindowHeight();

            return elementY >= windowY &&
            elementY - offsetY < windowY + windowHeight/2;
        };

        /**
         * Function to check if left border of an element is in the window
         *
         * @name Lazy#leftOfElementIsInWindow
         * @function
         * @param {jQuery} elem
         * @returns {Boolean}
         */
        self.leftOfElementIsInWindow = function (elem) {
            var elementX = self.getElementX(elem),
                offsetX  = self.oConf.offsetX,
                windowX = self.getWindowX(),
                windowWidth = self.getWindowWidth();

            return elementX >= windowX &&
            elementX - offsetX < windowX + windowWidth;
        };

        /**
         * Function to check if bottom border of an element is in the window
         *
         * @name Lazy#bottomOfElementIsInWindow
         * @function
         * @param {jQuery} elem
         * @returns {Boolean}
         */
        self.bottomOfElementIsInWindow = function (elem) {
            var elementY = self.getElementY(elem),
                elementHeight = self.getElementHeight(elem),
                offsetY  = self.oConf.offsetY,
                windowY = self.getWindowY(),
                windowHeight = self.getWindowHeight();

            return elementY + elementHeight > windowY &&
            elementY + elementHeight + offsetY <= windowY + windowHeight;
        };

        /**
         * Function to check if right of an element is in the window
         *
         * @name Lazy#rightOfElementIsInWindow
         * @function
         * @param {jQuery} elem
         * @returns {Boolean}
         */
        self.rightOfElementIsInWindow = function (elem) {
            var elementX = self.getElementX(elem),
                elementWidth = self.getElementWidth(elem),
                offsetX  = self.oConf.offsetX,
                windowX = self.getWindowX(),
                windowWidth = self.getWindowWidth();

            return elementX + elementWidth > windowX &&
            elementX + elementWidth + offsetX <= windowX + windowWidth;
        };

        /**
         * Function to check if left side of an element is on the left of the window
         * and right side of an element is on the right of the window
         *
         * @name Lazy#elementJutsOutOnLeftAndRight
         * @function
         * @param {jQuery} elem
         * @returns {Boolean}
         */
        self.elementJutsOutOnLeftAndRight = function (elem) {
            var elementY = self.getElementY(elem),
                elementHeight = self.getElementHeight(elem),
                windowY = self.getWindowY(),
                windowHeight = self.getWindowHeight();

            return elementY < windowY && (elementY + elementHeight) > (windowY + windowHeight);
        };

        /**
         * Function to check if top side of an element is over the top of the window
         * and bottom side of an element is under the bottom of the window
         *
         * @name Lazy#elementJutsOutOnTopAndBottom
         * @function
         * @param {jQuery} elem
         * @returns {Boolean}
         */
        self.elementJutsOutOnTopAndBottom = function (elem) {
            var elementX = self.getElementX(elem),
                elementWidth = self.getElementWidth(elem),
                windowX = self.getWindowX(),
                windowWidth = self.getWindowWidth();

            return elementX < windowX && (elementX + elementWidth) > (windowX + windowWidth);
        };

        /**
         * Functions to retrieve window Y position
         *
         * @name Lazy#getWindowY
         * @function
         * @returns {Number}
         */
        self.getWindowY = function () {
            return jQuery(window).scrollTop();
        };

        /**
         * Functions to retrieve window X positions
         *
         * @name Lazy#getWindowX
         * @function
         * @returns {Number}
         */
        self.getWindowX = function () {
            return jQuery(window).scrollLeft();
        };

        /**
         * Functions to retrieve window height
         *
         * @name Lazy#getWindowHeight
         * @function
         * @returns {Number}
         */
        self.getWindowHeight = function () {
            return self.getElementHeight(window);
        };

        /**
         * Functions to retrieve window width
         *
         * @name Lazy#getWindowWidth
         * @function
         * @returns {Number}
         */
        self.getWindowWidth = function () {
            return self.getElementWidth(window);
        };

        /**
         * Functions to retrieve element Y position
         *
         * @name Lazy#getElementY
         * @function
         * @returns {Number}
         */
        self.getElementY = function (elem) {
            return jQuery(elem).offset().top;
        };

        /**
         * Functions to retrieve element X position
         *
         * @name Lazy#getElementX
         * @function
         * @returns {Number}
         */
        self.getElementX = function (elem) {
            return jQuery(elem).offset().left;
        };

        /**
         * Functions to retrieve element height
         *
         * @name Lazy#getElementHeight
         * @function
         * @returns {Number}
         */
        self.getElementHeight = function (elem) {
            return jQuery(elem).height();
        };

        /**
         * Functions to retrieve element width
         *
         * @name Lazy#getElementWidth
         * @function
         * @return {Number}
         */
        self.getElementWidth = function (elem) {
            return jQuery(elem).width();
        };

        /**
         * Inject a javascript from data-src param
         *
         * @name BOM.Utils.injectScript
         * @param  jQueryObject $elem
         * @function
         * @return void
         */
        self.injectScript = function ($elem) {
            var url = $elem.data('src');

            if (url !== null && url !== '') {
                jQuery.ajax({
                    url: url,
                    dataType: 'script',
                    cache: true,
                    // context: $elem,
                    success: function () {
                        self.refreshUnloaded();
                    }
                });
            }
        };

        self.loadImage = function ($elem) {
            if ($elem.attr('data-srcset') !== 'undefined') {
                $elem.attr('srcset', $elem.attr('data-srcset'));
            }
            if ($elem.attr('data-src') !== 'undefined') {
                $elem.attr('src', $elem.attr('data-src'));
            }
        };

        /**
         * Retrieve JSON through ajax call.
         * The src is retrieved from data-src of the element.
         *
         * If the template JS is not already loaded, we don't do anything.
         *
         * @name Lazy#retrieveData
         * @function
         */
        self.retrieveData = function () {

            var $elem = jQuery(this),
                url = $elem.data('src'),
                dataLoaded = $elem.data('dataLoaded') || false,
                canLoadData = true,
                timeout =  parseInt($elem.data('timeout') || oConf.timeout, 10);

            if($elem.hasClass('loaded'))
            {
                return true;
            }

            if(!self.oConf.refreshLoaded){
                $elem.addClass('loaded');
                self.elements = self.elements.not($elem);
            }

            if(self.oConf.canReloadData === false && dataLoaded === true){
                canLoadData = false;
            }

            if ($elem.data('type') === 'script') {
                self.injectScript($elem);
                return true;
            }

            if ($elem.data('type') === oConf.callbackType) {
                oConf.onComplete($elem);
                return true;
            }

            if ($elem.is('img')) {
                self.loadImage($elem);
                return true;
            }

            if (url !== null && url !== '' && typeof url !== 'undefined' && canLoadData === true) {
                var qParams = {
                    url: url,
                    type: 'GET',
                    context: this,
                    dataType: $elem.data('tmpl') !== undefined ? ($elem.data('type') === 'jsonp' ? 'jsonp' : 'json') : 'html',
                    success: self.injectData,
                    error: oConf.onError || self.error,
                    beforeSend: function () {
                        $elem.addClass('loading');
                        if (oConf.onLoad && 'function' === typeof oConf.onLoad) {
                            oConf.onLoad($elem);
                        }
                    },
                    complete: function (data) {
                        $elem.removeClass('loading');
                        if (oConf.onComplete && 'function' === typeof oConf.onComplete) {
                            oConf.onComplete($elem);
                        }
                    }
                };

                if($elem.data('type') === 'jsonp'){
                    qParams.cache = true;
                    qParams.jsonpCallback = 'callback_' + (function(str){
                        var hash = 0;
                        if (str.length === 0) { return hash; }
                        for (var i = 0; i < str.length; i++) {
                            var char = str.charCodeAt(i);
                            hash = ((hash<<5)-hash)+char;
                            hash = hash & hash; // Convert to 32bit integer
                        }
                        return hash < 0 ? (-1*hash) : hash;
                    }(url));
                }

                self.doAjax(qParams, timeout);
                
                $elem.data('dataLoaded', true);
            }

            if(typeof self.oConf.onAppear === 'function'){
                self.oConf.onAppear($elem);
            }
            return true;
        };
        
        self.doAjax = function(qParams, timeout) {
            setTimeout(function() {
                jQuery.ajax(qParams);
            }, timeout);
        };

        /**
         * Inject data in the element with a template.
         * The template is retrieved from conf if defined or by id from data-tmpl of the element.
         *
         * @name Lazy#injectData
         * @function
         * @param {Object} data
         */
        self.injectData = function (data) {
            var $elem = jQuery(this);
            if (oConf.processData && 'function' === typeof oConf.processData) {
                data = oConf.processData(data, $elem);
            }

            var template = $elem.data('tmpl');

            if(typeof oConf.template === "function"){
                template = oConf.template($elem);
            }else if(undefined !== oConf.template){
                template = oConf.template;
            }

            if (data) {
                if (template !== undefined && template !== '') {
                    $elem.html(Templating.render(template, data));
                } else {
                    $elem.html(data);
                }
            }

            self.refreshUnloaded();

            if (oConf.afterSuccess && 'function' === typeof oConf.afterSuccess) {
                oConf.afterSuccess($elem, data);
            }

            $elem.trigger(oConf.lazyEventName);
        };

        /**
         * Logs error.
         *
         * @name Lazy#error
         * @function
         * @param {{status: string}} xhr
         */
        self.error = function (xhr) {
            if (typeof console !== 'undefined' && xhr.status > 200) {
                var errorMsg = 'Lazy: ' + xhr.status + ' error';
                if (typeof console.error === 'function') {
                    console.error(errorMsg);
                } else if (typeof console.log === 'function') {
                    console.log(errorMsg);
                }
            }
        };

        /**
         * Public
         * Enable lazy elements by selector.
         *
         * @name Lazy#enableElm
         * @param jQuery selector
         * @function
         */
        self.enableElm = function ($sel) {
            jQuery($sel).removeClass('loaded');
            self.refreshUnloaded();
        };

        /**
         * Public
         * Disable lazy elements by selector.
         *
         * @name Lazy#disableElm
         * @param jQuery selector
         * @function
         */
        self.disableElm = function ($sel) {
            jQuery($sel).addClass('loaded');
            self.refreshUnloaded();
        };

        if (oConf.init) {
            self.init();
        }

        return self;
    };

    Lazy.initModule = function(conf) {
        return new Lazy(conf || {});
    };

    return Lazy;
});